import os, datetime, subprocess, sys
from copy import deepcopy
import random
import numpy as np
import torch
import SimpleITK as sitk
from contextlib import contextmanager, redirect_stderr, redirect_stdout
from os import devnull
from typing import Optional, Union


@contextmanager
def suppress_stdout_stderr():
    """A context manager that redirects stdout and stderr to devnull"""
    with open(devnull, "w") as fnull:
        with redirect_stderr(fnull) as err, redirect_stdout(fnull) as out:
            yield (err, out)


def checkPatchDivisibility(patch_size: np.ndarray, number: Optional[int] = 16) -> bool:
    """
    This function checks the divisibility of a numpy array or integer for architectural integrity

    Args:
        patch_size (np.ndarray): The patch size for checking.
        number (Optional[int], optional): The number to check divisibility for. Defaults to 16.

    Returns:
        bool: If all elements of array are divisible or not, after taking 2D patches into account.
    """
    if isinstance(patch_size, int):
        patch_size_to_check = np.array(patch_size)
    else:
        patch_size_to_check = patch_size
    # for 2D, don't check divisibility of last dimension
    if patch_size_to_check[-1] == 1:
        patch_size_to_check = patch_size_to_check[:-1]
    # for 2D, don't check divisibility of first dimension
    elif patch_size_to_check[0] == 1:
        patch_size_to_check = patch_size_to_check[1:]
    if np.count_nonzero(np.remainder(patch_size_to_check, number)) > 0:
        return False

    # adding check to address https://github.com/mlcommons/GaNDLF/issues/53
    # there is quite possibly a better way to do this
    unique = np.unique(patch_size_to_check)
    if (unique.shape[0] == 1) and (unique[0] < number):
        return False
    return True


def get_date_time() -> str:
    """
    Get a well-parsed date string

    Returns:
        str: The date in format YYYY/MM/DD::HH:MM:SS
    """
    now = datetime.datetime.now().strftime("%Y/%m/%d::%H:%M:%S")
    return now


def get_unique_timestamp() -> str:
    """
    Get a well-parsed timestamp string to be used for unique filenames

    Returns:
        str: The date in format YYYYMMDD_HHMMSS
    """
    now = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
    return now


def get_filename_extension_sanitized(filename: str) -> str:
    """
    This function returns the extension of the filename with leading and trailing characters removed.
    Args:
        filename (str): The filename to be processed.
    Returns:
        str: The filename with extension removed.
    """
    _, ext = os.path.splitext(filename)
    # if .gz or .nii file is detected, always return .nii.gz
    if (ext == ".gz") or (ext == ".nii"):
        ext = ".nii.gz"
    return ext


def parse_version(version_string: str) -> int:
    """
    Parses version string, discards last identifier (NR/alpha/beta) and returns an integer for comparison.

    Args:
        version_string (str): The string to be parsed.

    Returns:
        int: The version number.
    """
    version_string_split = version_string.replace("-dev", "")
    version_string_split = version_string_split.split(".")
    if len(version_string_split) > 3:
        del version_string_split[-1]
    return int("".join(version_string_split))


def version_check(version_from_config: str, version_to_check: str) -> bool:
    """
    This function checks if the version of the config file is compatible with the version of the code.

    Args:
        version_from_config (str): The version of the config file.
        version_to_check (str): The version of the code or model to check.

    Returns:
        bool: If the version of the config file is compatible with the version of the code.
    """
    version_to_check_int = parse_version(version_to_check)
    min_ver = parse_version(version_from_config["minimum"])
    max_ver = parse_version(version_from_config["maximum"])
    if (min_ver > version_to_check_int) or (max_ver < version_to_check_int):
        sys.exit(
            "Incompatible version of GaNDLF detected ("
            + str(version_to_check_int)
            + ")"
        )

    return True


def checkPatchDimensions(patch_size: np.ndarray, numlay: int) -> int:
    """
    This function checks the divisibility of a numpy array or integer for architectural integrity

    Args:
        patch_size (np.ndarray): The patch size for checking.
        number (int, optional): The number to check divisibility for. Defaults to 16.

    Returns:
        int: Largest multiple of 2 (less than or equal to numlay) that each element of patch size is divisible by to yield int >= 2
    """
    if isinstance(patch_size, int):
        patch_size_to_check = np.array(patch_size)
    else:
        patch_size_to_check = patch_size
    # for 2D, don't check divisibility of last dimension
    if patch_size_to_check[-1] == 1:
        patch_size_to_check = patch_size_to_check[:-1]

    if all(
        [x >= 2 ** (numlay + 1) and x % 2**numlay == 0 for x in patch_size_to_check]
    ):
        return numlay
    else:
        # base2 = np.floor(np.log2(patch_size_to_check))
        base2 = np.array([getBase2(x) for x in patch_size_to_check])
        remain = patch_size_to_check / 2**base2  # check that at least 1

        layers = np.where(remain == 1, base2 - 1, base2)
        return int(np.min(layers))


def getBase2(num: int) -> int:
    """
    Compute the base 2 logarithm of a number.

    Args:
        num (int): the number

    Returns:
        int: the base 2 logarithm of the number
    """
    # helper for checkPatchDimensions (returns the largest multiple of 2 that num is evenly divisible by)
    base = 0
    while num % 2 == 0:
        num = num / 2
        base = base + 1
    return base


def get_array_from_image_or_tensor(
    input_tensor_or_image: Union[torch.Tensor, sitk.Image]
) -> np.ndarray:
    """
    This function returns the numpy array from a torch.Tensor or sitk.Image.

    Args:
        input_tensor_or_image (Union[torch.Tensor, sitk.Image]): The input tensor or image.

    Returns:
        np.ndarray: The numpy array.
    """
    assert isinstance(
        input_tensor_or_image, (torch.Tensor, sitk.Image, np.ndarray)
    ), "Input must be a torch.Tensor or sitk.Image or np.ndarray, but got " + str(
        type(input_tensor_or_image)
    )
    if isinstance(input_tensor_or_image, torch.Tensor):
        return input_tensor_or_image.detach().cpu().numpy()
    elif isinstance(input_tensor_or_image, sitk.Image):
        return sitk.GetArrayFromImage(input_tensor_or_image)
    elif isinstance(input_tensor_or_image, np.ndarray):
        return input_tensor_or_image


def set_determinism(seed: Optional[int] = 42) -> None:
    """
    This function sets the determinism for the random number generators.

    Args:
        seed (Optional[int], optional): The seed for the random number generators. Defaults to 42.
    """
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.use_deterministic_algorithms(True, warn_only=True)

    if torch.cuda.is_available():
        os.environ["CUBLAS_WORKSPACE_CONFIG"] = ":16:8"
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = True


def print_and_format_metrics(
    cohort_level_metrics: dict,
    sample_level_metrics: dict,
    metrics_dict_from_parameters: dict,
    mode: str,
    length_of_dataloader: int,
) -> dict:
    """
    This function prints and formats the metrics.

    Args:
        cohort_level_metrics (dict): The cohort level metrics calculated from the GANDLF.metrics.overall_stats function.
        sample_level_metrics (dict): The sample level metrics calculated from separate samples from the dataloader(s).
        metrics_dict_from_parameters (dict): The metrics dictionary to populate.
        mode (str): The mode of the metrics (train, val, test).
        length_of_dataloader (int): The length of the dataloader.

    Returns:
        dict: The metrics dictionary populated with the metrics.
    """

    def __update_metric_from_list_to_single_string(input_metrics_dict: dict) -> dict:
        """
        Helper function to update the metric from list to single string.

        Args:
            input_metrics_dict (dict): The input metrics dictionary.

        Returns:
            dict: The output metrics dictionary.
        """
        print(input_metrics_dict)
        output_metrics_dict = deepcopy(input_metrics_dict)
        for metric in input_metrics_dict.keys():
            if isinstance(input_metrics_dict[metric], list):
                output_metrics_dict[metric] = ("_").join(
                    str(input_metrics_dict[metric])
                    .replace("[", "")
                    .replace("]", "")
                    .replace(" ", "")
                    .split(",")
                )

        print(output_metrics_dict)
        return output_metrics_dict

    output_metrics_dict = deepcopy(cohort_level_metrics)
    for metric in metrics_dict_from_parameters:
        if isinstance(sample_level_metrics[metric], np.ndarray):
            to_print = (sample_level_metrics[metric] / length_of_dataloader).tolist()
        else:
            to_print = sample_level_metrics[metric] / length_of_dataloader
        output_metrics_dict[metric] = to_print
    for metric in output_metrics_dict.keys():
        print(
            "     Epoch Final   " + mode + " " + metric + " : ",
            output_metrics_dict[metric],
        )
    output_metrics_dict = __update_metric_from_list_to_single_string(
        output_metrics_dict
    )

    return output_metrics_dict


def define_average_type_key(params: dict, metric_name: str) -> str:
    """
    Determine the average type key from the metric config.

    Args:
        params (dict): The parameter dictionary containing training and data information.
        metric_name (str): The name of the metric.

    Returns:
        str: The average type key.
    """
    average_type_key = params["metrics"][metric_name].get("average", "macro")
    return average_type_key


def define_multidim_average_type_key(params: dict, metric_name: str) -> str:
    """
    Determine the multidimensional average type key from the metric config.

    Args:
        params (dict): The parameter dictionary containing training and data information.
        metric_name (str): The name of the metric.

    Returns:
        str: The multidimensional average type key.
    """
    average_type_key = params["metrics"][metric_name].get("multidim_average", "global")
    return average_type_key


def determine_classification_task_type(params: dict) -> str:
    """
    This function determines the classification task type from the parameters.

    Args:
        params (dict): The parameter dictionary containing training and data information.

    Returns:
        str: The classification task type (binary or multiclass).
    """
    task = "binary" if params["model"]["num_classes"] == 2 else "multiclass"
    return task


def get_git_hash() -> str:
    """
    Get the git hash of the current commit.

    Returns:
        str: The git hash of the current commit.
    """
    try:
        git_hash = (
            subprocess.check_output(["git", "rev-parse", "HEAD"], cwd=os.getcwd())
            .decode("ascii")
            .strip()
        )
    except (subprocess.CalledProcessError, FileNotFoundError):
        git_hash = "None"
    return git_hash



import os
import pathlib
from typing import Optional, Tuple

import numpy as np
import pandas as pd
import SimpleITK as sitk
import torch
from torch.utils.data import DataLoader
import torchio
from GANDLF.compute.loss_and_metric import get_loss_and_metrics
from GANDLF.compute.step import step
from GANDLF.data.post_process import global_postprocessing_dict
from GANDLF.utils import (
    get_date_time,
    get_filename_extension_sanitized,
    get_unique_timestamp,
    resample_image,
    reverse_one_hot,
    get_ground_truths_and_predictions_tensor,
    print_and_format_metrics,
)
from GANDLF.metrics import overall_stats
from tqdm import tqdm


def validate_network(
    model: torch.nn.Module,
    valid_dataloader: DataLoader,
    scheduler: object,
    params: dict,
    epoch: Optional[int] = 0,
    mode: Optional[str] = "validation",
) -> Tuple[float, dict]:
    """
    Function to validate a network for a single epoch.

    Args:
        model (torch.nn.Module): The model to process the input image with, it should support appropriate dimensions. if parameters["model"]["type"] == torch, this is a torch.model, otherwise this is OV exec_net.
        valid_dataloader (DataLoader): The dataloader for the validation epoch.
        scheduler (object): The scheduler to use for training.
        params (dict): The parameters passed by the user yaml.
        epoch (int, optional): The current epoch number. Defaults to 0.
        mode (str, optional): The mode of operation. Defaults to "validation".

    Returns:
        Tuple[float, dict]: The average validation loss and the average validation metrics.
    """
    print("*" * 20)
    print("Starting " + mode + " : ")
    print("*" * 20)
    # Initialize a few things
    total_epoch_valid_loss = 0
    total_epoch_valid_metric = {}
    average_epoch_valid_metric = {}

    for metric in params["metrics"]:
        if "per_label" in metric:
            total_epoch_valid_metric[metric] = []
        else:
            total_epoch_valid_metric[metric] = 0

    logits_list = []
    subject_id_list = []
    is_classification = params.get("problem_type") == "classification"
    calculate_overall_metrics = (
        (params["problem_type"] == "classification")
        or (params["problem_type"] == "regression")
    ) and mode == "validation"
    is_inference = mode == "inference"

    # automatic mixed precision - https://pytorch.org/docs/stable/amp.html
    if params["verbose"]:
        if params["model"]["amp"]:
            print("Using Automatic mixed precision", flush=True)

    if scheduler is None:
        current_output_dir = params["output_dir"]  # this is in inference mode
    else:  # this is useful for inference
        current_output_dir = os.path.join(params["output_dir"], "output_" + mode)

    if not (is_inference):
        current_output_dir = os.path.join(current_output_dir, str(epoch))

    pathlib.Path(current_output_dir).mkdir(parents=True, exist_ok=True)

    # Set the model to valid
    if params["model"]["type"] == "torch":
        model.eval()

    # # putting stuff in individual arrays for correlation analysis
    # all_targets = []
    # all_predics = []
    if params["medcam_enabled"] and params["model"]["type"] == "torch":
        model.enable_medcam()
        params["medcam_enabled"] = True

    if params["save_output"] or is_inference:
        if params["problem_type"] != "segmentation":
            outputToWrite = "Epoch,SubjectID,PredictedValue\n"
            file_to_write = os.path.join(current_output_dir, "output_predictions.csv")
            if os.path.exists(file_to_write):
                file_to_write = os.path.join(
                    current_output_dir,
                    "output_predictions_" + get_unique_timestamp() + ".csv",
                )

    # get ground truths for classification problem, validation set
    if calculate_overall_metrics:
        (
            ground_truth_array,
            predictions_array,
        ) = get_ground_truths_and_predictions_tensor(params, "validation_data")

    for batch_idx, (subject) in enumerate(
        tqdm(valid_dataloader, desc="Looping over " + mode + " data")
    ):
        if params["verbose"]:
            print("== Current subject:", subject["subject_id"], flush=True)

        # ensure spacing is always present in params and is always subject-specific
        params["subject_spacing"] = None
        if "spacing" in subject:
            params["subject_spacing"] = subject["spacing"]

        # constructing a new dict because torchio.GridSampler requires torchio.Subject,
        # which requires torchio.Image to be present in initial dict, which the loader does not provide
        subject_dict = {}
        label_ground_truth = None
        label_present = False
        # this is when we want the dataloader to pick up properties of GaNDLF's
        # DataLoader, such as pre-processing and augmentations, if appropriate
        if "label" in subject:
            if subject["label"] != ["NA"]:
                subject_dict["label"] = torchio.LabelMap(
                    path=subject["label"]["path"],
                    tensor=subject["label"]["data"].squeeze(0),
                    affine=subject["label"]["affine"].squeeze(0),
                )
                label_present = True
                label_ground_truth = subject_dict["label"]["data"]

        if "value_keys" in params:  # for regression/classification
            for key in params["value_keys"]:
                subject_dict["value_" + key] = subject[key]
                label_ground_truth = torch.cat(
                    [subject[key] for key in params["value_keys"]], dim=0
                )

        for key in params["channel_keys"]:
            subject_dict[key] = torchio.ScalarImage(
                path=subject[key]["path"],
                tensor=subject[key]["data"].squeeze(0),
                affine=subject[key]["affine"].squeeze(0),
            )

        # regression/classification problem AND label is present
        if (params["problem_type"] != "segmentation") and label_present:
            sampler = torchio.data.LabelSampler(params["patch_size"])
            tio_subject = torchio.Subject(subject_dict)
            generator = sampler(tio_subject, num_patches=params["q_samples_per_volume"])
            pred_output = 0
            for patch in generator:
                image = torch.cat(
                    [patch[key][torchio.DATA] for key in params["channel_keys"]], dim=0
                )
                valuesToPredict = torch.cat(
                    [patch["value_" + key] for key in params["value_keys"]], dim=0
                )
                image = image.unsqueeze(0)
                image = image.float().to(params["device"])
                ## special case for 2D
                assert params["model"]["type"] in [
                    "torch",
                    "openvino",
                ], "Model type not supported. Please only use 'torch' or 'openvino'."
                if image.shape[-1] == 1:
                    image = torch.squeeze(image, -1)
                if params["model"]["type"] == "torch":
                    pred_output += model(image)
                elif params["model"]["type"] == "openvino":
                    pred_output += torch.from_numpy(
                        model(
                            inputs={params["model"]["IO"][0][0]: image.cpu().numpy()}
                        )[params["model"]["IO"][1][0]]
                    )

            pred_output = pred_output.cpu() / params["q_samples_per_volume"]

            if is_inference and is_classification:
                logits_list.append(pred_output)
                subject_id_list.append(subject.get("subject_id")[0])

            if params["save_output"] or is_inference:
                # we divide by scaling factor here because we multiply by it during loss/metric calculation
                outputToWrite += (
                    str(epoch)
                    + ","
                    + subject["subject_id"][0]
                    + ","
                    + str(pred_output.cpu().max().item() / params["scaling_factor"])
                    + "\n"
                )
            final_loss, final_metric = get_loss_and_metrics(
                image, valuesToPredict, pred_output, params
            )

            if calculate_overall_metrics:
                predictions_array[batch_idx] = (
                    torch.argmax(pred_output[0], 0).cpu().item()
                )
            # # Non network validation related
            total_epoch_valid_loss += final_loss.detach().cpu().item()
            for metric in final_metric.keys():
                if isinstance(total_epoch_valid_metric[metric], list):
                    if len(total_epoch_valid_metric[metric]) == 0:
                        total_epoch_valid_metric[metric] = np.array(
                            final_metric[metric]
                        )
                    else:
                        total_epoch_valid_metric[metric] += np.array(
                            final_metric[metric]
                        )
                else:
                    total_epoch_valid_metric[metric] += final_metric[metric]

        else:  # for segmentation problems OR regression/classification when no label is present
            grid_sampler = torchio.inference.GridSampler(
                torchio.Subject(subject_dict),
                params["patch_size"],
                patch_overlap=params["inference_mechanism"]["patch_overlap"],
            )
            patch_loader = torch.utils.data.DataLoader(grid_sampler, batch_size=1)
            aggregator = torchio.inference.GridAggregator(
                grid_sampler,
                overlap_mode=params["inference_mechanism"]["grid_aggregator_overlap"],
            )

            if params["medcam_enabled"]:
                attention_map_aggregator = torchio.inference.GridAggregator(
                    grid_sampler,
                    overlap_mode=params["inference_mechanism"][
                        "grid_aggregator_overlap"
                    ],
                )

            output_prediction = 0  # this is used for regression/classification
            current_patch = 0
            for patches_batch in patch_loader:
                if params["verbose"]:
                    print(
                        "=== Current patch:",
                        current_patch,
                        ", time : ",
                        get_date_time(),
                        ", location :",
                        patches_batch[torchio.LOCATION],
                        flush=True,
                    )
                current_patch += 1
                image = (
                    torch.cat(
                        [
                            patches_batch[key][torchio.DATA]
                            for key in params["channel_keys"]
                        ],
                        dim=1,
                    )
                    .float()
                    .to(params["device"])
                )

                # calculate metrics if ground truth is present
                label = None
                if params["problem_type"] != "segmentation":
                    label = label_ground_truth
                elif "label" in patches_batch:
                    label = patches_batch["label"][torchio.DATA]

                if label is not None:
                    label = label.to(params["device"])
                    if params["verbose"]:
                        print(
                            "=== Validation shapes : label:",
                            label.shape,
                            ", image:",
                            image.shape,
                            flush=True,
                        )

                if is_inference:
                    result = step(model, image, None, params, train=False)
                else:
                    result = step(model, image, label, params, train=True)

                # get the current attention map and add it to its aggregator
                if params["medcam_enabled"]:
                    _, _, output, attention_map = result
                    attention_map_aggregator.add_batch(
                        attention_map, patches_batch[torchio.LOCATION]
                    )
                else:
                    _, _, output, _ = result

                if params["problem_type"] == "segmentation":
                    aggregator.add_batch(
                        output.detach().cpu(), patches_batch[torchio.LOCATION]
                    )
                else:
                    if torch.is_tensor(output):
                        # this probably needs customization for classification (majority voting or median, perhaps?)
                        output_prediction += output.detach().cpu()
                    else:
                        output_prediction += output

            # save outputs
            if params["problem_type"] == "segmentation":
                output_prediction = aggregator.get_output_tensor()
                output_prediction = output_prediction.unsqueeze(0)
                if params["save_output"]:
                    img_for_metadata = torchio.ScalarImage(
                        tensor=subject["1"]["data"].squeeze(0),
                        affine=subject["1"]["affine"].squeeze(0),
                    ).as_sitk()
                    pred_mask = output_prediction.numpy()
                    # perform postprocessing before reverse one-hot encoding here
                    for postprocessor in params["data_postprocessing"]:
                        for _class in range(0, params["model"]["num_classes"]):
                            pred_mask[0, _class, ...] = global_postprocessing_dict[
                                postprocessor
                            ](pred_mask[0, _class, ...], params)
                    # '0' because validation/testing dataloader always has batch size of '1'
                    pred_mask = reverse_one_hot(
                        pred_mask[0], params["model"]["class_list"]
                    )
                    pred_mask = np.swapaxes(pred_mask, 0, 2)

                    # perform postprocessing after reverse one-hot encoding here
                    for postprocessor in params[
                        "data_postprocessing_after_reverse_one_hot_encoding"
                    ]:
                        pred_mask = global_postprocessing_dict[postprocessor](
                            pred_mask, params
                        )

                    # if jpg detected, convert to 8-bit arrays
                    ext = get_filename_extension_sanitized(subject["1"]["path"][0])
                    if ext in [".jpg", ".jpeg", ".png"]:
                        pred_mask = pred_mask.astype(np.uint8)

                    ## special case for 2D
                    if image.shape[-1] > 1:
                        result_image = sitk.GetImageFromArray(pred_mask)
                    else:
                        result_image = sitk.GetImageFromArray(pred_mask.squeeze(0))
                    result_image.CopyInformation(img_for_metadata)

                    # this handles cases that need resampling/resizing
                    if "resample" in params["data_preprocessing"]:
                        result_image = resample_image(
                            result_image,
                            img_for_metadata.GetSpacing(),
                            interpolator=sitk.sitkNearestNeighbor,
                        )
                    # Create the subject directory if it doesn't exist in the
                    # current_output_dir directory
                    os.makedirs(
                        os.path.join(current_output_dir, "testing"), exist_ok=True
                    )
                    os.makedirs(
                        os.path.join(
                            current_output_dir, "testing", subject["subject_id"][0]
                        ),
                        exist_ok=True,
                    )

                    path_to_save = os.path.join(
                        current_output_dir,
                        "testing",
                        subject["subject_id"][0],
                        subject["subject_id"][0] + "_seg" + ext,
                    )

                    sitk.WriteImage(result_image, path_to_save)
            else:
                # final regression output
                output_prediction = output_prediction / len(patch_loader)
                if calculate_overall_metrics:
                    predictions_array[batch_idx] = (
                        torch.argmax(output_prediction[0], 0).cpu().item()
                    )
                if params["save_output"]:
                    outputToWrite += (
                        str(epoch)
                        + ","
                        + subject["subject_id"][0]
                        + ","
                        + str(output_prediction)
                        + "\n"
                    )

            # get the final attention map and save it
            if params["medcam_enabled"] and params["model"]["type"] == "torch":
                attention_map = attention_map_aggregator.get_output_tensor()
                for i, n in enumerate(attention_map):
                    model.save_attention_map(
                        n.squeeze(), raw_input=image[i].squeeze(-1)
                    )

            output_prediction = output_prediction.squeeze(-1)
            if is_inference and is_classification:
                logits_list.append(output_prediction)
                subject_id_list.append(subject.get("subject_id")[0])

            # we cast to float32 because float16 was causing nan
            if label_ground_truth is not None:
                # this is for RGB label
                if label_ground_truth.shape[0] == 3:
                    label_ground_truth = label_ground_truth[0, ...].unsqueeze(0)
                # we always want the ground truth to be in the same format as the prediction
                label_ground_truth = label_ground_truth.unsqueeze(0)
                if label_ground_truth.shape[-1] == 1:
                    label_ground_truth = label_ground_truth.squeeze(-1)
                final_loss, final_metric = get_loss_and_metrics(
                    image,
                    label_ground_truth,
                    output_prediction.to(torch.float32),
                    params,
                )
                if params["verbose"]:
                    print(
                        "Full image " + mode + ":: Loss: ",
                        final_loss,
                        "; Metric: ",
                        final_metric,
                        flush=True,
                    )

                # # Non network validing related
                # loss.cpu().data.item()
                total_epoch_valid_loss += final_loss.cpu().item()
                for metric in final_metric.keys():
                    if isinstance(total_epoch_valid_metric[metric], list):
                        if len(total_epoch_valid_metric[metric]) == 0:
                            total_epoch_valid_metric[metric] = np.array(
                                final_metric[metric]
                            )
                        else:
                            total_epoch_valid_metric[metric] += np.array(
                                final_metric[metric]
                            )
                    else:
                        total_epoch_valid_metric[metric] += final_metric[metric]

        if label_ground_truth is not None:
            if params["verbose"]:
                # For printing information at halftime during an epoch
                if ((batch_idx + 1) % (len(valid_dataloader) / 2) == 0) and (
                    (batch_idx + 1) < len(valid_dataloader)
                ):
                    print(
                        "\nHalf-Epoch Average " + mode + " loss : ",
                        total_epoch_valid_loss / (batch_idx + 1),
                    )
                    for metric in params["metrics"]:
                        if isinstance(total_epoch_valid_metric[metric], np.ndarray):
                            to_print = (
                                total_epoch_valid_metric[metric] / (batch_idx + 1)
                            ).tolist()
                        else:
                            to_print = total_epoch_valid_metric[metric] / (
                                batch_idx + 1
                            )
                        print(
                            "Half-Epoch Average " + mode + " " + metric + " : ",
                            to_print,
                        )

    if params["medcam_enabled"] and params["model"]["type"] == "torch":
        model.disable_medcam()
        params["medcam_enabled"] = False

    if label_ground_truth is not None:
        average_epoch_valid_loss = total_epoch_valid_loss / len(valid_dataloader)
        print("     Epoch Final   " + mode + " loss : ", average_epoch_valid_loss)
        # get overall stats for classification
        if calculate_overall_metrics:
            average_epoch_valid_metric = overall_stats(
                predictions_array, ground_truth_array, params
            )
        average_epoch_valid_metric = print_and_format_metrics(
            average_epoch_valid_metric,
            total_epoch_valid_metric,
            params["metrics"],
            mode,
            len(valid_dataloader),
        )

    else:
        average_epoch_valid_loss, average_epoch_valid_metric = 0, {}

    if scheduler is not None:
        if params["scheduler"]["type"] in [
            "reduce_on_plateau",
            "reduce-on-plateau",
            "plateau",
            "reduceonplateau",
        ]:
            scheduler.step(average_epoch_valid_loss)
        else:
            scheduler.step()

    # write the predictions, if appropriate
    if params["save_output"]:
        if is_inference and is_classification and logits_list:
            class_list = [str(c) for c in params["model"]["class_list"]]
            logit_tensor = torch.cat(logits_list)
            current_fold_dir = params["current_fold_dir"]
            logit_tensor = logit_tensor.detach().cpu().numpy()
            columns = ["SubjectID"] + class_list
            logits_df = pd.DataFrame(columns=columns)
            logits_df.SubjectID = subject_id_list
            logits_df[class_list] = logit_tensor

            logits_file = os.path.join(current_fold_dir, "logits.csv")
            if os.path.isfile(logits_file):
                logits_file = os.path.join(
                    current_fold_dir, "logits_" + get_unique_timestamp() + ".csv"
                )
            logits_df.to_csv(logits_file, index=False, sep=",")

        if "value_keys" in params:
            file = open(file_to_write, "w")
            file.write(outputToWrite)
            file.close()

    return average_epoch_valid_loss, average_epoch_valid_metric



import os, pathlib, pytest
from pytest import fixture

from .test_full import (
    prerequisites_hook_download_data,
    prerequisites_constructTrainingCSV,
)


def pytest_addoption(parser):
    parser.addoption(
        "--device", action="store", default="cpu", help="device option: cpu or cuda"
    )


@fixture()
def device(request):
    return request.config.getoption("--device")


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item, call):
    # execute all other hooks to obtain the report object
    outcome = yield
    rep = outcome.get_result()

    # we only look at actual failing test calls, not setup/teardown
    if rep.when == "call" and rep.failed:
        log_filename = os.path.join(
            pathlib.Path(__file__).parent.absolute(), "failures.log"
        )
        mode = "a" if os.path.exists(log_filename) else "w"
        with open(log_filename, mode) as f:
            f.write(rep.longreprtext + "\n")


def pytest_sessionstart(session):
    """
    This hook is executed before the pytest session starts.
    """
    prerequisites_hook_download_data()
    prerequisites_constructTrainingCSV()



import sys
import yaml
from typing import Optional
from pprint import pprint
import pandas as pd
from tqdm import tqdm
import torch
import torchio
import SimpleITK as sitk
import numpy as np

from GANDLF.config_manager import ConfigManager
from GANDLF.utils import find_problem_type_from_parameters, one_hot
from GANDLF.metrics import (
    overall_stats,
    structural_similarity_index,
    mean_squared_error,
    peak_signal_noise_ratio,
    mean_squared_log_error,
    mean_absolute_error,
    ncc_mean,
    ncc_std,
    ncc_max,
    ncc_min,
)
from GANDLF.losses.segmentation import dice
from GANDLF.metrics.segmentation import (
    _calculator_generic_all_surface_distances,
    _calculator_sensitivity_specificity,
    _calculator_jaccard,
)


def generate_metrics_dict(
    input_csv: str, config: str, outputfile: Optional[str] = None
) -> dict:
    """
    This function generates metrics from the input csv and the config.

    Args:
        input_csv (str): The input CSV.
        config (str): The input yaml config.
        outputfile (str, optional): The output file to save the metrics. Defaults to None.

    Returns:
        dict: The metrics dictionary.
    """
    input_df = pd.read_csv(input_csv)

    # check required headers in a case insensitive manner
    headers = {}
    required_columns = ["subjectid", "prediction", "target"]
    for col, _ in input_df.items():
        col_lower = col.lower()
        for column_to_check in required_columns:
            if column_to_check == col_lower:
                headers[column_to_check] = col
        if col_lower == "mask":
            headers["mask"] = col
    for column in required_columns:
        assert column in headers, f"The input csv should have a column named {column}"

    overall_stats_dict = {}
    parameters = ConfigManager(config)
    problem_type = parameters.get("problem_type", None)
    problem_type = (
        find_problem_type_from_parameters(parameters)
        if problem_type is None
        else problem_type
    )
    parameters["problem_type"] = problem_type

    if problem_type == "regression" or problem_type == "classification":
        parameters["model"]["num_classes"] = len(parameters["model"]["class_list"])
        predictions_tensor = torch.from_numpy(
            input_df[headers["prediction"]].to_numpy().ravel()
        )
        labels_tensor = torch.from_numpy(input_df[headers["target"]].to_numpy().ravel())
        overall_stats_dict = overall_stats(
            predictions_tensor, labels_tensor, parameters
        )

    elif problem_type == "segmentation":
        # read images and then calculate metrics
        class_list = parameters["model"]["class_list"]
        for _, row in tqdm(input_df.iterrows(), total=input_df.shape[0]):
            current_subject_id = row[headers["subjectid"]]
            overall_stats_dict[current_subject_id] = {}
            label_image = torchio.LabelMap(row[headers["target"]])
            pred_image = torchio.LabelMap(row[headers["prediction"]])
            label_tensor = label_image.data
            pred_tensor = pred_image.data
            spacing = label_image.spacing
            if label_tensor.data.shape[-1] == 1:
                spacing = spacing[0:2]
            # add dimension for batch
            parameters["subject_spacing"] = torch.Tensor(spacing).unsqueeze(0)
            label_tensor = label_tensor.unsqueeze(0)
            pred_tensor = pred_tensor.unsqueeze(0)

            # one hot encode with batch_size = 1
            label_image_one_hot = one_hot(label_tensor, class_list)
            pred_image_one_hot = one_hot(pred_tensor, class_list)

            for class_index, _ in enumerate(class_list):
                current_target = label_image_one_hot[:, class_index, ...].unsqueeze(0)
                current_prediction = pred_image_one_hot[:, class_index, ...].unsqueeze(
                    0
                )
                overall_stats_dict[current_subject_id][str(class_index)] = {}
                # this is inconsequential, since one_hot will ensure that the classes are present
                parameters["model"]["class_list"] = [1]
                parameters["model"]["num_classes"] = 1
                overall_stats_dict[current_subject_id][str(class_index)]["dice"] = dice(
                    current_prediction, current_target
                ).item()
                nsd, hd100, hd95 = _calculator_generic_all_surface_distances(
                    current_prediction, current_target, parameters
                )
                overall_stats_dict[current_subject_id][str(class_index)][
                    "nsd"
                ] = nsd.item()
                overall_stats_dict[current_subject_id][str(class_index)][
                    "hd100"
                ] = hd100.item()
                overall_stats_dict[current_subject_id][str(class_index)][
                    "hd95"
                ] = hd95.item()

                (s, p) = _calculator_sensitivity_specificity(
                    current_prediction, current_target, parameters
                )
                overall_stats_dict[current_subject_id][str(class_index)][
                    "sensitivity"
                ] = s.item()
                overall_stats_dict[current_subject_id][str(class_index)][
                    "specificity"
                ] = p.item()
                overall_stats_dict[current_subject_id][
                    "jaccard_" + str(class_index)
                ] = _calculator_jaccard(
                    current_prediction, current_target, parameters
                ).item()
                current_target_image = sitk.GetImageFromArray(
                    current_target[0, 0, ...].long()
                )
                current_prediction_image = sitk.GetImageFromArray(
                    current_prediction[0, 0, ...].long()
                )
                label_overlap_filter = sitk.LabelOverlapMeasuresImageFilter()
                label_overlap_filter.Execute(
                    current_target_image, current_prediction_image
                )
                # overall_stats_dict[current_subject_id][
                #     "falseDiscoveryRate_" + str(class_index)
                # ] = label_overlap_filter.GetFalseDiscoveryRate()
                overall_stats_dict[current_subject_id][
                    "falseNegativeError_" + str(class_index)
                ] = label_overlap_filter.GetFalseNegativeError()
                overall_stats_dict[current_subject_id][
                    "falsePositiveError_" + str(class_index)
                ] = label_overlap_filter.GetFalsePositiveError()
                overall_stats_dict[current_subject_id][
                    "meanOverlap_" + str(class_index)
                ] = label_overlap_filter.GetMeanOverlap()
                overall_stats_dict[current_subject_id][
                    "unionOverlap_" + str(class_index)
                ] = label_overlap_filter.GetUnionOverlap()
                overall_stats_dict[current_subject_id][
                    "volumeSimilarity_" + str(class_index)
                ] = label_overlap_filter.GetVolumeSimilarity()

    elif problem_type == "synthesis":

        def __fix_2d_tensor(input_tensor):
            """
            This function checks for 2d images and change the shape to [B, C, H, W]

            Args:
                input_tensor (torch.Tensor): The input tensor.

            Returns:
                torch.Tensor: The output tensor in the format that torchmetrics expects.
            """
            if input_tensor.shape[-1] == 1:
                return input_tensor.squeeze(-1).unsqueeze(0)
            else:
                return input_tensor

        def __percentile_clip(
            input_tensor: torch.Tensor,
            reference_tensor: torch.Tensor = None,
            p_min: Optional[float] = 0.5,
            p_max: Optional[float] = 99.5,
            strictlyPositive: Optional[bool] = True,
        ):
            """
            Normalizes a tensor based on percentiles. Clips values below and above the percentile.
            Percentiles for normalization can come from another tensor.

            Args:
                input_tensor (torch.Tensor): Tensor to be normalized based on the data from the reference_tensor. If reference_tensor is None, the percentiles from this tensor will be used.
                reference_tensor (torch.Tensor, optional): The tensor used for obtaining the percentiles.
                p_min (float, optional): Lower end percentile. Defaults to 0.5.
                p_max (float, optional): Upper end percentile. Defaults to 99.5.
                strictlyPositive (bool, optional): Ensures that really all values are above 0 before normalization. Defaults to True.

            Returns:
                torch.Tensor: The input_tensor normalized based on the percentiles of the reference tensor.
            """
            reference_tensor = (
                input_tensor if reference_tensor is None else reference_tensor
            )
            v_min, v_max = np.percentile(
                reference_tensor, [p_min, p_max]
            )  # get p_min percentile and p_max percentile

            # set lower bound to be 0 if strictlyPositive is enabled
            v_min = max(v_min, 0.0) if strictlyPositive else v_min
            output_tensor = np.clip(
                input_tensor, v_min, v_max
            )  # clip values to percentiles from reference_tensor
            output_tensor = (output_tensor - v_min) / (
                v_max - v_min
            )  # normalizes values to [0;1]
            return output_tensor

        for _, row in tqdm(input_df.iterrows(), total=input_df.shape[0]):
            current_subject_id = row[headers["subjectid"]]
            overall_stats_dict[current_subject_id] = {}
            target_image = __fix_2d_tensor(
                torchio.ScalarImage(row[headers["target"]]).data
            )
            pred_image = __fix_2d_tensor(
                torchio.ScalarImage(row[headers["prediction"]]).data
            )
            # if "mask" is not in the row, we assume that the whole image is the mask
            # always cast to byte tensor
            mask = (
                __fix_2d_tensor(torchio.LabelMap(row[headers["mask"]]).data)
                if "mask" in row
                else torch.from_numpy(
                    np.ones(target_image.numpy().shape, dtype=np.uint8)
                )
            ).byte()

            # Get Infill region (we really are only interested in the infill region)
            output_infill = (pred_image * mask).float()
            gt_image_infill = (target_image * mask).float()

            # Normalize to [0;1] based on GT (otherwise MSE will depend on the image intensity range)
            normalize = parameters.get("normalize", True)
            if normalize:
                reference_tensor = (
                    target_image * ~mask
                )  # use all the tissue that is not masked for normalization
                gt_image_infill = __percentile_clip(
                    gt_image_infill,
                    reference_tensor=reference_tensor,
                    p_min=0.5,
                    p_max=99.5,
                    strictlyPositive=True,
                )
                output_infill = __percentile_clip(
                    output_infill,
                    reference_tensor=reference_tensor,
                    p_min=0.5,
                    p_max=99.5,
                    strictlyPositive=True,
                )

            overall_stats_dict[current_subject_id][
                "ssim"
            ] = structural_similarity_index(output_infill, gt_image_infill, mask).item()

            # ncc metrics
            compute_ncc = parameters.get("compute_ncc", True)
            if compute_ncc:
                overall_stats_dict[current_subject_id]["ncc_mean"] = ncc_mean(
                    output_infill, gt_image_infill
                )
                overall_stats_dict[current_subject_id]["ncc_std"] = ncc_std(
                    output_infill, gt_image_infill
                )
                overall_stats_dict[current_subject_id]["ncc_max"] = ncc_max(
                    output_infill, gt_image_infill
                )
                overall_stats_dict[current_subject_id]["ncc_min"] = ncc_min(
                    output_infill, gt_image_infill
                )

            # only voxels that are to be inferred (-> flat array)
            # these are required for mse, psnr, etc.
            gt_image_infill = gt_image_infill[mask]
            output_infill = output_infill[mask]

            overall_stats_dict[current_subject_id]["mse"] = mean_squared_error(
                output_infill, gt_image_infill
            ).item()

            overall_stats_dict[current_subject_id]["msle"] = mean_squared_log_error(
                output_infill, gt_image_infill
            ).item()

            overall_stats_dict[current_subject_id]["mae"] = mean_absolute_error(
                output_infill, gt_image_infill
            ).item()

            # torchmetrics PSNR using "max"
            overall_stats_dict[current_subject_id]["psnr"] = peak_signal_noise_ratio(
                output_infill, gt_image_infill
            ).item()

            # same as above but with epsilon for robustness
            overall_stats_dict[current_subject_id][
                "psnr_eps"
            ] = peak_signal_noise_ratio(
                output_infill, gt_image_infill, epsilon=sys.float_info.epsilon
            ).item()

            # only use fix data range to [0;1] if the data was normalized before
            if normalize:
                # torchmetrics PSNR but with fixed data range of 0 to 1
                overall_stats_dict[current_subject_id][
                    "psnr_01"
                ] = peak_signal_noise_ratio(
                    output_infill, gt_image_infill, data_range=(0, 1)
                ).item()

                # same as above but with epsilon for robustness
                overall_stats_dict[current_subject_id][
                    "psnr_01_eps"
                ] = peak_signal_noise_ratio(
                    output_infill,
                    gt_image_infill,
                    data_range=(0, 1),
                    epsilon=sys.float_info.epsilon,
                ).item()

    pprint(overall_stats_dict)
    if outputfile is not None:
        with open(outputfile, "w") as outfile:
            yaml.dump(overall_stats_dict, outfile)



import numpy as np

from .crop_zero_planes import CropExternalZeroplanes
from .non_zero_normalize import NonZeroNormalizeOnMaskedRegion
from .threshold_and_clip import threshold_transform, clip_transform
from .normalize_rgb import (
    normalize_by_val_transform,
    normalize_imagenet_transform,
    normalize_standardize_transform,
    normalize_div_by_255_transform,
)
from .template_matching import histogram_matching, stain_normalizer
from .resample_minimum import Resample_Minimum
from .rgb_conversion import rgba2rgb_transform, rgb2rgba_transform

from torchio.transforms import (
    ZNormalization,
    ToCanonical,
    Crop,
    CropOrPad,
    Resize,
    Resample,
    Compose,
    RescaleIntensity,
)


def generic_3d_check(patch_size):
    """
    This function reads the value from the configuration and returns an appropriate tuple for torchio to ingest.

    Args:
        patch_size (Union[list, tuple, array]): The generic list/tuple/array to check.

    Returns:
        tuple: The tuple to be ingested by torchio.
    """
    patch_size_new = np.array(patch_size)
    if len(patch_size) == 2:
        patch_size_new = tuple(np.append(np.array(patch_size), 1))

    return patch_size_new


def positive_voxel_mask(image):
    return image > 0


def nonzero_voxel_mask(image):
    return image != 0


def to_canonical_transform(parameters=None):
    return ToCanonical()


def crop_transform(patch_size):
    return Crop(generic_3d_check(patch_size))


def centercrop_transform(patch_size):
    return CropOrPad(target_shape=generic_3d_check(patch_size))


def rescale_transform(parameters=None):
    if parameters is None:
        parameters = {}
    # get defaults from torchio
    rescaler = RescaleIntensity()
    rescaler.out_min_max = parameters.get("out_min_max", rescaler.out_min_max)
    rescaler.percentiles = parameters.get("percentiles", rescaler.percentiles)
    rescaler.in_min_max = parameters.get("in_min_max", None)
    return rescaler


# defining dict for pre-processing - key is the string and the value is the transform object
global_preprocessing_dict = {
    "to_canonical": to_canonical_transform,
    "threshold": threshold_transform,
    "clip": clip_transform,
    "clamp": clip_transform,
    "crop_external_zero_planes": CropExternalZeroplanes,
    "crop": crop_transform,
    "centercrop": centercrop_transform,
    "normalize_by_val": normalize_by_val_transform,
    "normalize_imagenet": normalize_imagenet_transform(),
    "normalize_standardize": normalize_standardize_transform(),
    "normalize_div_by_255": normalize_div_by_255_transform(),
    "normalize": ZNormalization(),
    "normalize_positive": ZNormalization(masking_method=positive_voxel_mask),
    "normalize_nonZero": ZNormalization(masking_method=nonzero_voxel_mask),
    "normalize_nonzero": ZNormalization(masking_method=nonzero_voxel_mask),
    "normalize_nonZero_masked": NonZeroNormalizeOnMaskedRegion(),
    "normalize_nonzero_masked": NonZeroNormalizeOnMaskedRegion(),
    "rescale": rescale_transform,
    "rgba2rgb": rgba2rgb_transform,
    "rgbatorgb": rgba2rgb_transform,
    "rgba_to_rgb": rgba2rgb_transform,
    "rgb2rgba": rgb2rgba_transform,
    "rgbtorgba": rgb2rgba_transform,
    "rgb_to_rgba": rgb2rgba_transform,
    "histogram_matching": histogram_matching,
    "stain_normalizer": stain_normalizer,
}


def get_transforms_for_preprocessing(
    parameters, current_transformations, train_mode, apply_zero_crop
):
    """
    This function gets the pre-processing transformations from the parameters.

    Args:
        parameters (dict): The parameters dictionary.
        current_transformations (list): The current transformations list.
        train_mode (bool): Whether the data is in train mode or not.
        apply_zero_crop (bool): Whether to apply zero crop or not.

    Returns:
        list: The list of pre-processing transformations.
    """

    preprocessing_params_dict = parameters["data_preprocessing"]
    # first, we want to do thresholding, followed by clipping, if it is present - required for inference as well
    normalize_to_apply = None
    if not (preprocessing_params_dict is None):
        # go through preprocessing in the order they are specified
        for preprocess in preprocessing_params_dict:
            preprocess_lower = preprocess.lower()
            # special check for resize and resample
            if preprocess_lower == "resize_patch":
                resize_values = generic_3d_check(preprocessing_params_dict[preprocess])
                current_transformations.append(Resize(resize_values))
            elif preprocess_lower == "resample":
                if "resolution" in preprocessing_params_dict[preprocess]:
                    # Need to take a look here
                    resample_values = generic_3d_check(
                        preprocessing_params_dict[preprocess]["resolution"]
                    )
                    current_transformations.append(Resample(resample_values))
            elif preprocess_lower in ["resample_minimum", "resample_min"]:
                if "resolution" in preprocessing_params_dict[preprocess]:
                    resample_values = generic_3d_check(
                        preprocessing_params_dict[preprocess]["resolution"]
                    )
                    current_transformations.append(Resample_Minimum(resample_values))
            # special check for histogram_matching
            elif preprocess_lower == "histogram_matching":
                if preprocessing_params_dict[preprocess] is not False:
                    current_transformations.append(
                        global_preprocessing_dict[preprocess_lower](
                            preprocessing_params_dict[preprocess]
                        )
                    )
            # special check for stain_normalizer
            elif preprocess_lower == "stain_normalizer":
                if normalize_to_apply is None:
                    normalize_to_apply = global_preprocessing_dict[preprocess_lower](
                        preprocessing_params_dict[preprocess]
                    )
            # normalize should be applied at the end
            elif "normalize" in preprocess_lower:
                if normalize_to_apply is None:
                    normalize_to_apply = global_preprocessing_dict[preprocess_lower]
            # preprocessing routines that we only want for training
            elif preprocess_lower in ["crop_external_zero_planes"]:
                if train_mode or apply_zero_crop:
                    current_transformations.append(
                        global_preprocessing_dict["crop_external_zero_planes"](
                            patch_size=parameters["patch_size"]
                        )
                    )
            # everything else is taken in the order passed by user
            elif preprocess_lower in global_preprocessing_dict:
                current_transformations.append(
                    global_preprocessing_dict[preprocess_lower](
                        preprocessing_params_dict[preprocess]
                    )
                )

    # normalization type is applied at the end
    if normalize_to_apply is not None:
        current_transformations.append(normalize_to_apply)

    # compose the transformations
    transforms_to_apply = None
    if current_transformations:
        transforms_to_apply = Compose(current_transformations)

    return transforms_to_apply



import sys
import torch.nn as nn
import torch.nn.functional as F
from collections import OrderedDict
import math
import numpy as np

from .modelBase import ModelBase


class _MBConv1(nn.Sequential):
    def __init__(
        self,
        num_in_feats,
        num_out_feats,
        kernel_size,
        stride,
        output_size,
        reduction,  # reduction factor for squeeze excitation
        Norm,
        Conv,
        Pool,
    ):
        """
        Defines a Single MBConv block in the EfficientNet model.

        Args:
            num_in_feats: Number of input features.
            num_out_feats: Number of output features.
            kernel_size: Size of the convolutional kernel.
            stride: Stride of the convolutional layer.
            output_size: Size of the output tensor.
            reduction: Reduction factor for squeeze excitation.
            Norm: Normalization layer for the MBConv block.
            Conv: Convolutional layer for the MBConv block.
            Pool: Pooling layer for the squeeze excitation block.

        Returns:
            Output tensor after passing through the MBConv block.

        """
        super().__init__()

        # depthwise conv -> batch norm -> swish
        # SE
        # conv -> batch norm

        self.add_module(
            "depthconv1",
            Conv(
                num_in_feats,
                num_in_feats,
                groups=num_in_feats,
                kernel_size=kernel_size,
                stride=stride,
                padding=int((kernel_size - 1) / 2),
                bias=False,
            ),
        )
        self.add_module("norm1", Norm(num_in_feats))
        self.add_module("silu", nn.SiLU(inplace=True))

        # ADD SQUEEZE EXCITATION; no change in dimensions
        self.add_module(
            "squeeze", _SqueezeExcitation(num_in_feats, reduction, output_size, Pool)
        )

        self.add_module(
            "conv2",
            Conv(
                num_in_feats,
                num_out_feats,
                kernel_size=1,
                stride=1,
                padding=int((kernel_size - 1) / 2),
                bias=False,
            ),
        )

        self.add_module("norm2", Norm(num_out_feats))

    def forward(self, x):
        """
        Sequentially passes the input tensor through depthwise convolution -> batch normalization -> swish activation ->
        squeeze excitation -> convolution -> batch normalization and returns the output tensor.

        Args:
            x (torch.Tensor): Input tensor for the MBConv block.

        Returns:
            out (torch.Tensor): Output tensor after passing through the MBConv block.
        """
        out = self.depthconv1(x)
        out = self.norm1(out)
        out = self.silu(out)

        out = self.squeeze(out)

        out = self.conv2(out)
        out = self.norm2(out)

        return out


class _MBConv6(nn.Sequential):
    def __init__(
        self,
        num_in_feats,
        num_out_feats,
        kernel_size,
        stride,
        output_size,
        Norm,
        Conv,
        Pool,
        reduction,  # reduction factor for squeeze excitation
    ):
        """
        A class representing a single MobileNetV3 block with expansion factor 6. Refer to https://arxiv.org/abs/1905.02244 for more.

        Args:
            num_in_feats (int): The number of input features/channels.
            num_out_feats (int): The number of output features/channels.
            kernel_size (int): The size of the kernel/convolutional filter.
            stride (int): The stride of the convolutional operation.
            output_size (int): The output size of the block.
            Norm (torch.nn.Module): A normalization layer to be used in the block.
            Conv (torch.nn.Module): A convolutional layer to be used in the block.
            Pool (torch.nn.Module): A pooling layer to be used in the squeeze-excitation block.
            reduction (float): The reduction factor for the squeeze-excitation block.
        """
        super().__init__()

        # conv -> batch norm --> swish
        # depthwise conv -> batch norm --> swish
        # SE
        # conv -> batch norm

        self.add_module(
            "conv1",
            Conv(
                num_in_feats,
                6 * num_in_feats,
                kernel_size=1,
                stride=1,
                padding=0,
                bias=False,
            ),
        )
        self.add_module("norm1", Norm(6 * num_in_feats))
        self.add_module("silu", nn.SiLU(inplace=True))

        self.add_module(
            "depthconv1",
            Conv(
                6 * num_in_feats,
                6 * num_in_feats,
                kernel_size=kernel_size,
                stride=stride,
                padding=int((kernel_size - 1) / 2),
                bias=False,
            ),
        )
        self.add_module("norm2", Norm(6 * num_in_feats))

        # ADD SQUEEZE EXCITATION; no change in dimensions
        self.add_module(
            "squeeze",
            _SqueezeExcitation(6 * num_in_feats, reduction, output_size, Pool),
        )

        self.add_module(
            "conv2",
            Conv(
                6 * num_in_feats,
                num_out_feats,
                kernel_size=1,
                stride=1,
                padding=0,
                bias=False,
            ),
        )

        self.add_module("norm3", Norm(num_out_feats))

    def forward(self, x):
        """
        Performs a forward pass through the model.

        Args:
            x (torch.Tensor): The input tensor to the model.

        Returns:
            torch.Tensor: The output tensor from the model.
        """
        out = self.conv1(x)
        out = self.norm1(out)
        out = self.silu(out)

        out = self.depthconv1(out)
        out = self.norm2(out)
        out = self.silu(out)

        out = self.squeeze(out)

        out = self.conv2(out)
        out = self.norm3(out)

        return out


class _SqueezeExcitation(nn.Sequential):
    def __init__(
        self,
        num_in_feats,
        reduction,  # reduce dimension to int(num_in_feats / reduction)
        output_size,
        Pool,
    ):
        """
        Squeeze-and-Excitation block that recalibrates channel-wise feature responses

        Args:
            num_in_feats (int): Number of input features
            reduction (int): Reduction factor for number of hidden features
            output_size (tuple): Tuple containing the size of the output tensor
            Pool (nn.Module): Pooling operation
        """
        super().__init__()

        # global avg pool
        # FC --> ReLU
        # FC --> Sigmoid

        self.add_module("FC1", nn.Linear(num_in_feats, int(num_in_feats / reduction)))
        self.add_module("relu", nn.ReLU(inplace=True))

        self.add_module("FC2", nn.Linear(int(num_in_feats / reduction), num_in_feats))
        self.add_module("sigmoid", nn.Sigmoid())
        self.Pool = Pool
        self.output_size = output_size

    def forward(self, x):
        """
        Forward pass for the Squeeze-and-Excitation block

        Args:
            x (torch.Tensor): Input tensor of shape [batch_size, in_channels, H, W]

        Returns:
            torch.Tensor: Output tensor of the same shape as the input tensor
        """
        out = self.Pool(self.output_size)(x).view(x.size(0), -1)
        out = self.FC1(out)
        out = self.relu(out)
        out = self.FC2(out)
        out = self.sigmoid(out)

        dims = list(x.size())
        dims[2:] = self.output_size
        return x * out.view(dims).expand_as(x)


DEFAULT_BLOCKS = [
    {
        "block_type": _MBConv1,
        "kernel": 3,
        "output_size": 16,
        "num_layers": 1,
        "stride": 1,
    },
    {
        "block_type": _MBConv6,
        "kernel": 3,
        "output_size": 24,
        "num_layers": 2,
        "stride": 2,
    },
    {
        "block_type": _MBConv6,
        "kernel": 5,
        "output_size": 40,
        "num_layers": 2,
        "stride": 2,
    },
    {
        "block_type": _MBConv6,
        "kernel": 3,
        "output_size": 80,
        "num_layers": 3,
        "stride": 2,
    },
    {
        "block_type": _MBConv6,
        "kernel": 5,
        "output_size": 112,
        "num_layers": 3,
        "stride": 1,
    },
    {
        "block_type": _MBConv6,
        "kernel": 5,
        "output_size": 192,
        "num_layers": 4,
        "stride": 2,
    },
    {
        "block_type": _MBConv6,
        "kernel": 3,
        "output_size": 320,
        "num_layers": 1,
        "stride": 1,
    },
]


def checkPatchDimensions(patch_size, numlay):
    """
    Check if patch dimensions are divisible by 2^numlay.

    Args:
        patch_size (tuple or int): Tuple of integers or a single integer representing the patch size.
        numlay (int): Number of layers.

    Returns:
        int: The largest number of layers that are divisible by 2^numlay.
    """
    patch_size_to_check = np.atleast_1d(patch_size)  # Convert to array if not already
    if patch_size_to_check[-1] == 1:
        patch_size_to_check = patch_size_to_check[:-1]

    if all(patch_size_to_check >= 2**numlay) and all(
        patch_size_to_check % 2**numlay == 0
    ):
        return numlay
    else:
        base2 = np.array(
            [np.log2(x).astype(int) for x in patch_size_to_check]
        )  # get largest possible number of layers for each dim
        return int(np.min(base2))


def num_channels(default_chan, width_factor, divisor):
    """
    Compute the number of channels closest to default_chan * width_factor that is divisible by divisor.

    Args:
        default_chan (int): Default number of channels.
        width_factor (float): Width factor.
        divisor (int): Divisor.

    Returns:
        int: The number of channels closest to default_chan * width_factor that is divisible by divisor.
    """
    default_chan *= width_factor
    new_out = int(default_chan + divisor / 2) // divisor * divisor
    new_out = max(new_out, divisor)

    if new_out < 0.9 * default_chan:
        new_out += divisor

    return int(new_out)


def num_layers(default_lay, depth_factor):
    """
    Compute the number of layers closest to default_lay * depth_factor.

    Args:
        default_lay (int): Default number of layers.
        depth_factor (float): Depth factor.

    Returns:
        int: The number of layers closest to default_lay * depth_factor.
    """
    return int(math.ceil(default_lay * depth_factor))


class EfficientNet(ModelBase):
    """
    EfficientNet model with customizable scaling of depth and width.

    Args:
        parameters (dict): A dictionary of configuration parameters for the model.
        scale_params (dict) - A dictionary defining scaling of depth and width for the model.
    """

    def __init__(self, parameters: dict, scale_params):  # how to scale depth and width
        super(EfficientNet, self).__init__(parameters)

        # check/define defaults

        if self.n_dimensions == 2:
            self.output_size = (1, 1)
        elif self.n_dimensions == 3:
            self.output_size = (1, 1, 1)
        else:
            sys.exit("Only 2D or 3D convolutions are supported.")
        if self.Norm is None:
            sys.stderr.write(
                "Warning: efficientnet is not defined without a normalization layer"
            )
            self.Norm = self.BatchNorm

        patch_check = checkPatchDimensions(parameters["patch_size"], numlay=5)
        self.DEFAULT_BLOCKS = DEFAULT_BLOCKS

        common_msg = "The patch size is not large enough for desired number of layers. It is expected that each dimension of the patch size is divisible by 2^i, where i is in a integer greater than or equal to 2."

        assert not (patch_check != 5 and patch_check <= 1), common_msg

        if patch_check != 5 and patch_check >= 2:
            downsamp = np.where(
                np.array([x["stride"] == 2 for x in self.DEFAULT_BLOCKS])
            )[0][patch_check - 1]
            self.DEFAULT_BLOCKS = self.DEFAULT_BLOCKS[:downsamp]
            print(common_msg + " Only the first %d layers will run." % patch_check)

        num_out_channels = num_channels(32, scale_params["width"], 8)
        # first convolution: 3x3 conv stride 2, norm, swish
        self.features = [
            (
                "conv1",
                self.Conv(
                    in_channels=self.n_channels,
                    out_channels=num_out_channels,
                    kernel_size=3,
                    stride=2,
                    padding=1,
                    bias=False,
                ),
            ),
            ("norm1", self.Norm(num_out_channels)),
            ("swish1", nn.SiLU(inplace=True)),
        ]
        self.features = nn.Sequential(OrderedDict(self.features))

        for i, block in enumerate(self.DEFAULT_BLOCKS):
            for i_lay in range(
                0, num_layers(block["num_layers"], scale_params["depth"])
            ):
                temp_out = num_channels(block["output_size"], scale_params["width"], 8)
                layer = block["block_type"](
                    num_in_feats=num_out_channels,
                    num_out_feats=temp_out,
                    kernel_size=block["kernel"],
                    Norm=self.Norm,
                    Conv=self.Conv,
                    Pool=self.AdaptiveAvgPool,
                    stride=(block["stride"] - 1) * (i_lay == 0) + 1,
                    output_size=self.output_size,
                    reduction=4,
                )
                self.features.add_module("block%d-layer%d" % (i, i_lay), layer)
                num_out_channels = temp_out

        # final convolution : conv 1x1, pooling, fc

        final_conv = [
            (
                "conv2",
                self.Conv(
                    in_channels=num_out_channels,
                    out_channels=num_channels(1280, scale_params["width"], 8),
                    kernel_size=1,
                    stride=1,
                    padding=0,
                    bias=False,
                ),
            ),
            ("norm2", self.Norm(num_channels(1280, scale_params["width"], 8))),
            ("swish2", nn.SiLU(inplace=True)),
        ]

        self.features.add_module("final_conv", nn.Sequential(OrderedDict(final_conv)))

        self.classifier = nn.Sequential(
            nn.Dropout(p=scale_params["dropout"], inplace=True),
            nn.Linear(num_channels(1280, scale_params["width"], 8), self.n_classes),
        )

        # define starting weights
        for m in self.modules():
            if isinstance(m, self.Conv):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(m, self.Norm):
                if m.weight is not None:
                    m.weight.data.fill_(1)
                if m.bias is not None:
                    m.bias.data.zero_()
            elif isinstance(m, nn.Linear):
                nn.init.constant_(m.bias, 0)

    def forward(self, x):
        """
        Applies the model's layers to the input tensor and produces an output tensor.

        Args:
            x (torch.Tensor): Input tensor to the model.

        Returns:
            torch.Tensor: Output tensor produced by the model.
        """
        # Pass the input tensor through the feature extraction layers
        features = self.features(x)

        # Apply adaptive average pooling to the output tensor from the previous step
        out = self.AdaptiveAvgPool(self.output_size)(features).view(
            features.size(0), -1
        )

        # Pass the tensor through the classifier layers
        out = self.classifier(out)

        # Apply the final convolution layer if not None
        if not self.final_convolution_layer is None:
            if self.final_convolution_layer == F.softmax:
                out = self.final_convolution_layer(out, dim=1)
            else:
                out = self.final_convolution_layer(out)

        return out


def efficientnetB0(parameters):
    return EfficientNet(
        parameters, scale_params={"width": 1.0, "depth": 1.0, "dropout": 0.2}
    )


def efficientnetB1(parameters):
    return EfficientNet(
        parameters, scale_params={"width": 1.0, "depth": 1.1, "dropout": 0.2}
    )


def efficientnetB2(parameters):
    return EfficientNet(
        parameters, scale_params={"width": 1.1, "depth": 1.2, "dropout": 0.3}
    )


def efficientnetB3(parameters):
    return EfficientNet(
        parameters, scale_params={"width": 1.2, "depth": 1.4, "dropout": 0.3}
    )


def efficientnetB4(parameters):
    return EfficientNet(
        parameters, scale_params={"width": 1.4, "depth": 1.8, "dropout": 0.4}
    )


def efficientnetB5(parameters):
    return EfficientNet(
        parameters, scale_params={"width": 1.6, "depth": 2.2, "dropout": 0.4}
    )


def efficientnetB6(parameters):
    return EfficientNet(
        parameters, scale_params={"width": 1.8, "depth": 2.6, "dropout": 0.5}
    )


def efficientnetB7(parameters):
    return EfficientNet(
        parameters, scale_params={"width": 2.0, "depth": 3.1, "dropout": 0.5}
    )



import torch.nn as nn


class DownsamplingModule(nn.Module):
    def __init__(
        self,
        input_channels,
        output_channels,
        conv=nn.Conv2d,
        conv_kwargs=None,
        norm=nn.BatchNorm2d,
        norm_kwargs=None,
        act=nn.LeakyReLU,
        act_kwargs=None,
    ):
        """
        Initializes a downsampling module with convolution, batch normalization, and activation layers.

        Args:
            input_channels (int): Number of input channels
            output_channels (int): Number of output channels
            conv (nn.Module): Convolution layer
            conv_kwargs (dict): Convolution layer keyword arguments
            norm (nn.Module): Normalization layer
            norm_kwargs (dict): Normalization layer keyword arguments
            act (nn.Module): Activation function layer
            act_kwargs (dict): Activation function layer keyword arguments
        """
        super(DownsamplingModule, self).__init__()
        if conv_kwargs is None:
            conv_kwargs = {"kernel_size": 3, "stride": 2, "padding": 1, "bias": True}
        if norm_kwargs is None:
            norm_kwargs = {
                "eps": 1e-5,
                "affine": True,
                "momentum": 0.1,
                "track_running_stats": True,
            }
        if act_kwargs is None:
            act_kwargs = {"negative_slope": 1e-2, "inplace": True}

        self.in_0 = norm(output_channels, **norm_kwargs)

        self.conv0 = conv(input_channels, output_channels, **conv_kwargs)

        self.act = act(**act_kwargs)

    def forward(self, x):
        """
        Applies a downsampling operation to the input tensor.

        [input -- > in --> lrelu --> ConvDS --> output]

        Args:
        x (torch.Tensor): Input tensor of shape (batch_size, channels, height, width)

        Returns:
            torch.Tensor: The output tensor, of shape (batch_size, output_channels, height // 2, width // 2).
        """
        x = self.act(self.in_0(self.conv0(x)))

        return x
